/*
 *  Copyright (C) 2024  stsp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <sys/prctl.h>
#include <searpc.h>
#include "test-object.h"
#include "priv.h"
#include "utilities.h"
#include "emu.h"
#include "fsrpcdefs.h"
#include "fssvc.h"
#include "fssvc_priv.h"

static SearpcClient *clnt;
static int sock_rx;
static int exited;

static void svc_ex(void *arg)
{
    error("fssvc failed, exiting\n");
    exited++;
    leavedos(35);
}

/*
 * arg: rpc_client->arg. Normally a socket number
 * fcall_str: the JSON data stream generated by Searpc
 * fcall_len: the length of `fcall_str`
 * ret_len: place to get the length of the returned json data stream
 * Returns: A newly allocated string stores the JSON data stream
 */
static char *transport_callback(void *arg, const char *fcall_str,
        size_t fcall_len, size_t *ret_len)
{
    char buf[4096];
    int sock = (int)(uintptr_t)arg;
    ssize_t sd = send(sock, fcall_str, fcall_len, MSG_DONTWAIT);
    if (sd <= 0)
        return NULL;
    sd = recv(sock, buf, sizeof(buf), 0);
    if (sd <= 0)
        return NULL;
    *ret_len = sd;
    return g_strndup(buf, sd);
}

int fssvc_init(plist_idx_t plist_idx, setattr_t setattr_cb,
    getattr_t getattr_cb)
{
    int socks[2];
    int transp[2];
    pid_t pid;
    int err;
    pshared_sem_t svc_sem;

    err = socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, socks);
    if (err) {
        perror("socketpair()");
        return -1;
    }
    err = socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, transp);
    if (err) {
        perror("socketpair()");
        goto err0;
    }
    err = pshared_sem_init(&svc_sem, 0);
    if (err)
        goto err1;
    pid = fork();
    switch (pid) {
        case -1:
            perror("fork()");
            goto err2;
        case 0:
            /* child */
            close(socks[0]);
            close(transp[0]);
            err = priv_drop();
            if (err) {
                pshared_sem_post(svc_sem);
                pshared_sem_destroy(&svc_sem);
                _exit(1);
            }
            setsid();
            prctl(PR_SET_PDEATHSIG, SIGQUIT);
            err = fsrpc_srv_init(transp[1], socks[1], plist_idx, setattr_cb,
                    getattr_cb);
            pshared_sem_post(svc_sem);
            pshared_sem_destroy(&svc_sem);
            if (err) {
                fprintf(stderr, "fs service failed\n");
                _exit(1);
            }
            fsrpc_svc_run();
            _exit(1);  // not reached
            break;
    }

    close(socks[1]);
    close(transp[1]);
    pshared_sem_wait(svc_sem);
    pshared_sem_destroy(&svc_sem);

    clnt = searpc_client_new();
    clnt->send = transport_callback;
    clnt->arg = (void *)(uintptr_t)transp[0];
    sigchld_register_handler(pid, svc_ex, NULL);
    sock_rx = socks[0];
    return 0;

err2:
    pshared_sem_destroy(&svc_sem);
err1:
    close(transp[0]);
    close(transp[1]);
err0:
    close(socks[0]);
    close(socks[1]);
    return -1;
}

static int recv_fd(int sock)
{
    struct msghdr msg;
    struct cmsghdr *cmsghdr;
    struct iovec iov;
    ssize_t nbytes;
    int *p;
    char buf[CMSG_SPACE(sizeof(int))], c[16];
    struct cmsghdr *cmsgp;

    iov.iov_base = &c;
    iov.iov_len = sizeof(c);
    cmsghdr = (struct cmsghdr *)buf;
    cmsghdr->cmsg_len = CMSG_LEN(sizeof(int));
    cmsghdr->cmsg_level = SOL_SOCKET;
    cmsghdr->cmsg_type = SCM_RIGHTS;
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    msg.msg_control = cmsghdr;
    msg.msg_controllen = CMSG_LEN(sizeof(int));
    msg.msg_flags = 0;

    nbytes = recvmsg(sock, &msg, 0);
    if (nbytes == -1)
        return -1;

    cmsgp = CMSG_FIRSTHDR(&msg);
    p = (int *)CMSG_DATA(cmsgp);
    if (!p)
        return -1;
    return *p;
}

static void bad_rpc(const char *func, const char *msg)
{
    fprintf(stderr, "RPC failure: %s: %s\n", func, msg);
    if (!exited) {
        exited++;
        leavedos(5);
    }
}

#define BAD_RPC(msg) do { \
    bad_rpc(__FUNCTION__, msg); \
    return -1; \
} while (0)
#define CHECK_RPC(st) do { \
    if (st) { \
        BAD_RPC(st->message); \
        g_error_free(st); \
    } \
} while (0)

int fssvc_add_path(const char *path)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "add_path_1",
                                  &error, 1,
                                  "string", path);
    CHECK_RPC(error);
    return ret;
}

int fssvc_add_path_ex(const char *path)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "add_path_ex_1",
                                  &error, 1,
                                  "string", path);
    CHECK_RPC(error);
    return ret;
}

int fssvc_add_path_list(const char *list)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "add_path_list_1",
                                  &error, 1,
                                  "string", list);
    CHECK_RPC(error);
    return ret;
}

int fssvc_seal(void)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "seal_1", &error, 0);
    CHECK_RPC(error);
    return ret;
}

static void bad_call(int err, const char *func)
{
    dosemu_error("%s returned %s, exiting\n", func, strerror(err));
    if (!exited) {
        exited++;
        leavedos(6);
    }
}

#define CHECK_RET(_r) do { \
    TestObject *r = TEST_OBJECT(_r); \
    if ((r)->ret < 0) { \
        int _rv; \
        if ((r)->args_err) \
            bad_call((r)->args_err, __FUNCTION__); \
        else \
            errno = (r)->errn; \
        _rv = (r)->ret; \
        g_object_unref(_r); \
        return _rv; \
    } \
} while (0)

int fssvc_open(int id, const char *path, int flags)
{
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "open_1", TEST_OBJECT_TYPE,
                                     &error, 3,
                                     "int", id, "string", path,
                                     "int", flags);
    CHECK_RPC(error);
    CHECK_RET(ret);
    g_object_unref(ret);
    return recv_fd(sock_rx);
}

int fssvc_creat(int id, const char *path, int flags, mode_t mode)
{
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "creat_1", TEST_OBJECT_TYPE,
                                     &error, 4,
                                     "int", id, "string", path,
                                     "int", flags, "int", mode);
    CHECK_RPC(error);
    CHECK_RET(ret);
    g_object_unref(ret);
    return recv_fd(sock_rx);
}

#define RPC_EPILOG() \
    CHECK_RPC(error); \
    CHECK_RET(ret); \
    rv = TEST_OBJECT(ret)->ret; \
    g_object_unref(ret); \
    return rv

int fssvc_unlink(int id, const char *path)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "unlink_1", TEST_OBJECT_TYPE,
                                     &error, 2,
                                     "int", id, "string", path);
    RPC_EPILOG();
}

int fssvc_setxattr(int id, const char *path, int attr)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "setxattr_1", TEST_OBJECT_TYPE,
                                     &error, 3,
                                     "int", id, "string", path, "int", attr);
    RPC_EPILOG();
}

int fssvc_getxattr(int id, const char *path)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "getxattr_1", TEST_OBJECT_TYPE,
                                     &error, 2,
                                     "int", id, "string", path);
    RPC_EPILOG();
}

int fssvc_rename(int id1, const char *path1, int id2, const char *path2)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "rename_1", TEST_OBJECT_TYPE,
                                     &error, 4,
                                     "int", id1, "string", path1,
                                     "int", id2, "string", path2);
    RPC_EPILOG();
}

int fssvc_mkdir(int id, const char *path, mode_t mode)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "mkdir_1", TEST_OBJECT_TYPE,
                                     &error, 3,
                                     "int", id, "string", path, "int", mode);
    RPC_EPILOG();
}

int fssvc_rmdir(int id, const char *path)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "rmdir_1", TEST_OBJECT_TYPE,
                                     &error, 2,
                                     "int", id, "string", path);
    RPC_EPILOG();
}

int fssvc_utime(int id, const char *path, time_t atime, time_t mtime)
{
    int rv;
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "utime_1", TEST_OBJECT_TYPE,
                                     &error, 4,
                                     "int", id, "string", path,
                                     /* searpc passes 64bit ints via ptrs */
                                     "int64", &atime, "int64", &mtime);
    RPC_EPILOG();
}

int fssvc_path_ok(int id, const char *path)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "path_ok_1",
                                  &error, 2,
                                  "int", id, "string", path);
    CHECK_RPC(error);
    return ret;
}

int fssvc_exit(void)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "exit_1", &error, 0);
    searpc_client_free(clnt);
    if (!in_leavedos)
        CHECK_RPC(error);
    return ret;
}

int fssvc_shm_open(const char *name, int oflag, mode_t mode)
{
    GObject* ret;
    GError *error = NULL;
    ret = searpc_client_call__object(clnt, "shm_open_1", TEST_OBJECT_TYPE,
                                     &error, 3,
                                     "string", name, "int", oflag,
                                     "int", mode);
    CHECK_RPC(error);
    CHECK_RET(ret);
    g_object_unref(ret);
    return recv_fd(sock_rx);
}

int fssvc_shm_unlink(const char *name)
{
    int ret;
    GError *error = NULL;
    ret = searpc_client_call__int(clnt, "shm_unlink_1",
                                  &error, 1,
                                  "string", name);
    CHECK_RPC(error);
    return ret;
}
